"""Main file for CLI."""
import sys
import argparse
from typing import Any, List, Optional

import shtab

import redbrick
from redbrick.cli.command import (
    CLIConfigController,
    CLIInitController,
    CLICloneController,
    CLIInfoController,
    CLIExportController,
    CLIUploadController,
    CLIIReportController,
)
from redbrick.cli.cli_base import CLIInterface
from redbrick.utils.logging import log_error, logger


class CLIController(CLIInterface):
    """Main CLI Controller."""

    def __init__(self, command: argparse._SubParsersAction) -> None:
        """Initialize CLI command parsers."""
        self.config = CLIConfigController(
            command.add_parser(
                self.CONFIG,
                help="Setup the credentials for your CLI.",
                description="Setup the credentials for your CLI.",
            )
        )
        self.init = CLIInitController(
            command.add_parser(
                self.INIT,
                help="Create a new project",
                description="""
Create a new project. We recommend creating a new directory and naming it after your project,
initializing your project within the new directory.

```bash
$ mkdir new-project
$ cd new-project
$ redbrick init
```
            """,
            )
        )
        self.clone = CLICloneController(
            command.add_parser(
                self.CLONE,
                help="Clone an existing remote project to local",
                description="""
The project will be cloned to a local directory named after your `project name`.
                """,
            )
        )
        self.info = CLIInfoController(
            command.add_parser(
                self.INFO,
                help="Get a project's information",
                description="Get a project's information",
            )
        )
        self.export = CLIExportController(
            command.add_parser(
                self.EXPORT,
                help="Export data for a project",
                description="Export data for a project",
            )
        )
        self.upload = CLIUploadController(
            command.add_parser(
                self.UPLOAD,
                help="Upload files to a project",
                description="Upload files to a project",
            )
        )
        self.report = CLIIReportController(
            command.add_parser(
                self.REPORT,
                help="Generate an audit report for a project",
                description="""
Generate an audit report for a project. Exports a JSON file containing all actions & events
associated with every task, including:

- Who annotated the task
- Who uploaded the data
- Who reviewed the task
- and more.
""",
            )
        )

    def handle_command(self, args: argparse.Namespace) -> None:
        """CLI command main handler."""
        if args.command == self.CONFIG:
            self.config.handler(args)
        elif args.command == self.INIT:
            self.init.handler(args)
        elif args.command == self.CLONE:
            self.clone.handler(args)
        elif args.command == self.INFO:
            self.info.handler(args)
        elif args.command == self.EXPORT:
            self.export.handler(args)
        elif args.command == self.UPLOAD:
            self.upload.handler(args)
        elif args.command == self.REPORT:
            self.report.handler(args)
        else:
            raise argparse.ArgumentError(None, "")


def cli_parser(only_parser: bool = True) -> Any:
    """Initialize argument parser."""
    parser = argparse.ArgumentParser(
        description="The RedBrick CLI offers a simple interface to quickly import and "
        + "export your images & annotations, and perform other high-level actions."
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"v{redbrick.__version__}"
    )
    cli = CLIController(parser.add_subparsers(title="Commands", dest="command"))

    shtab.add_argument_to(parser, "--completion")

    if only_parser:
        return parser

    return parser, cli


def cli_main(argv: Optional[List[str]] = None) -> None:
    """CLI main handler."""
    parser: argparse.ArgumentParser
    cli: CLIController

    parser, cli = cli_parser(False)

    try:
        args = parser.parse_args(argv if argv is not None else sys.argv[1:])
        logger.debug(args)
    except KeyboardInterrupt:
        logger.warning("User interrupted")
    except argparse.ArgumentError as error:
        logger.warning(str(error))
        parser.print_help()
    else:
        try:
            cli.handle_command(args)
        except KeyboardInterrupt:
            logger.warning("User interrupted")
        except argparse.ArgumentError as error:
            log_error(error)
            parser.print_usage()


if __name__ == "__main__":
    cli_main()



"""Abstract interface to upload."""
import json
from typing import List, Dict, Optional, Any

import aiohttp

from redbrick.common.client import RBClient
from redbrick.common.upload import UploadControllerInterface


class UploadRepo(UploadControllerInterface):
    """Handle communication with backend relating to uploads."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    async def create_datapoint_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels_data: Optional[str],
        labels_map: Optional[List[Dict]] = None,
        series_info: Optional[List[Dict]] = None,
        meta_data: Optional[str] = None,
        is_ground_truth: bool = False,
        pre_assign: Optional[Dict] = None,
    ) -> Dict:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """
        # pylint: disable=too-many-locals
        query_string = """
            mutation createDatapointSDK(
                $orgId: UUID!
                $projectId: UUID!
                $items: [String!]!
                $name: String!
                $storageId: UUID!
                $labelsData: String
                $labelsMap: [LabelMapInput!]
                $seriesInfo: [SeriesInfoInput!]
                $metaData: String
                $isGroundTruth: Boolean!
                $preAssign: String
            ) {
                createDatapoint(
                    orgId: $orgId
                    projectId: $projectId
                    items: $items
                    name: $name
                    storageId: $storageId
                    labelsData: $labelsData
                    labelsMap: $labelsMap
                    seriesInfo: $seriesInfo
                    metaData: $metaData
                    isGroundTruth: $isGroundTruth
                    preAssign: $preAssign
                ) {
                    taskId
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "items": items,
            "name": name,
            "storageId": storage_id,
            "labelsData": labels_data,
            "labelsMap": labels_map,
            "seriesInfo": series_info,
            "metaData": meta_data,
            "isGroundTruth": is_ground_truth,
            "preAssign": json.dumps(pre_assign, separators=(",", ":")),
        }
        response = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )
        return response.get("createDatapoint", {})

    async def update_items_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        task_id: str,
        items: List[str],
        series_info: Optional[List[Dict]] = None,
    ) -> Dict:
        """Update items in a datapoint."""
        query_string = """
            mutation updateTaskItems(
                $orgId: UUID!
                $projectId: UUID!
                $storageId: UUID!
                $taskId: UUID!
                $items: [String!]!
                $seriesInfo: [SeriesInfoInput!]
            ) {
                updateTaskItems(
                    orgId: $orgId
                    projectId: $projectId
                    storageId: $storageId
                    taskId: $taskId
                    items: $items
                    seriesInfo: $seriesInfo
                ) {
                    ok
                    message
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "storageId": storage_id,
            "taskId": task_id,
            "items": items,
            "seriesInfo": series_info,
        }
        response = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )
        return response.get("updateTaskItems", {})

    def items_upload_presign(
        self, org_id: str, project_id: str, files: List[str], file_type: List[str]
    ) -> List[Dict[Any, Any]]:
        """Return presigned URLs to upload files."""
        query_string = """
            query itemsUploadPresign(
                $orgId:UUID!,
                $projectId: UUID!,
                $files: [String]!,
                $fileType:[String]!
            ){
                itemsUploadPresign(
                    orgId:$orgId,
                    projectId: $projectId,
                    files:$files,
                    fileType:$fileType
                ) {
                    items {
                        presignedUrl,
                        filePath,
                        fileName
                    }
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "files": files,
            "fileType": file_type,
        }
        result = self.client.execute_query(query_string, query_variables)
        assert isinstance(result["itemsUploadPresign"]["items"], list)
        return result["itemsUploadPresign"]["items"]

    def delete_tasks(self, org_id: str, project_id: str, task_ids: List[str]) -> bool:
        """Delete tasks in a project."""
        query_string = """
        mutation deleteTasksSDK($orgId: UUID!, $projectId: UUID!, $taskIds: [UUID!]) {
            deleteTasks(
                orgId: $orgId
                projectId: $projectId
                taskIds: $taskIds
            ) {
                ok
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskIds": task_ids,
        }

        result: Dict[str, Dict] = self.client.execute_query(
            query_string, query_variables
        )

        return (result.get("deleteTasks", {}) or {}).get("ok", False)

    def delete_tasks_by_name(
        self, org_id: str, project_id: str, task_names: List[str]
    ) -> bool:
        """Delete tasks in a project by task names."""
        query_string = """
        mutation deleteTasksNamesSDK($orgId: UUID!, $projectId: UUID!, $taskNames: [String!]) {
            deleteTasks(
                orgId: $orgId
                projectId: $projectId
                taskNames: $taskNames
            ) {
                ok
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskNames": task_names,
        }

        result: Dict[str, Dict] = self.client.execute_query(
            query_string, query_variables
        )

        return (result.get("deleteTasks", {}) or {}).get("ok", False)

    async def generate_items_list(
        self,
        aio_client: aiohttp.ClientSession,
        files: List[str],
        import_type: str,
        as_study: bool = False,
        windows: bool = False,
    ) -> str:
        """Generate direct upload items list."""
        query_string = """
            query generateItemsList(
                $importType: ImportType!
                $files: [String]!
                $groupedByStudy: Boolean!
                $windows: Boolean
            ) {
                generateItemsList(
                    importType: $importType
                    files: $files
                    groupedByStudy: $groupedByStudy
                    windows: $windows
                )
            }
        """

        query_variables = {
            "importType": import_type,
            "files": files,
            "groupedByStudy": as_study,
            "windows": windows,
        }
        result = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )
        items_list: str = result["generateItemsList"]
        return items_list

    async def validate_and_convert_to_import_format(
        self,
        aio_client: aiohttp.ClientSession,
        original: str,
        convert: Optional[bool] = None,
        storage_id: Optional[str] = None,
    ) -> Dict:
        """Validate and convert tasks format."""
        query_string = """
        query validateAndConvertToImportFormat(
            $original: String!
            $convert: Boolean
            $storageId: UUID
        ) {
            validateAndConvertToImportFormat(
                original: $original
                convert: $convert
                storageId: $storageId
            ) {
                isValid
                error
                converted
            }
        }
        """
        query_variables = {
            "original": original,
            "convert": convert,
            "storageId": storage_id,
        }

        result: Dict[str, Dict] = await self.client.execute_query_async(
            aio_client, query_string, query_variables
        )

        return result.get("validateAndConvertToImportFormat", {}) or {}

    def import_tasks_from_workspace(
        self,
        org_id: str,
        project_id: str,
        source_project_id: str,
        task_search: List[Dict],
        with_labels: bool = False,
    ) -> Dict:
        """Import tasks from another project in the same workspace."""
        query_string = """
            mutation importTasksFromWorkspace(
                $orgId: UUID!
                $projectId: UUID!
                $sourceProjectId: UUID!
                $tasks: [TaskMetaDataInput!]!
                $withLabels: Boolean
            ) {
                importTasksFromWorkspace(
                    orgId: $orgId
                    projectId: $projectId
                    sourceProjectId: $sourceProjectId
                    tasks: $tasks
                    withLabels: $withLabels
                ) {
                    ok
                    message
                }
            }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "sourceProjectId": source_project_id,
            "tasks": task_search,
            "withLabels": with_labels,
        }
        result = self.client.execute_query(query_string, query_variables)
        return result.get("importTasksFromProject", {}) or {}



"""Logging functions."""
import os
from typing import Union
import logging

from rich.logging import RichHandler
from rich import pretty, traceback


debug_mode = bool(os.environ.get("REDBRICK_SDK_DEBUG"))
pretty.install(overflow="fold")
traceback.install(
    word_wrap=True, show_locals=debug_mode, max_frames=10 if debug_mode else 5
)

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[
        RichHandler(
            level=logging.DEBUG if debug_mode else logging.INFO,
            show_path=debug_mode,
            enable_link_path=False,
            markup=True,
        )
    ],
)

logger = logging.getLogger("redbrick")
logger.setLevel(logging.DEBUG if debug_mode else logging.INFO)


def log_error(text: Union[str, Exception]) -> None:
    """Log errors."""
    if debug_mode:
        if isinstance(text, str):
            raise ValueError(text)
        raise ValueError(str(text)) from text
    logger.error(text)



"""Constants."""


MAX_CONCURRENCY = 30
MAX_FILE_BATCH_SIZE = 5
MAX_RETRY_ATTEMPTS = 3

DEFAULT_URL = "https://api.redbrickai.com"

PEERLESS_ERRORS = (
    KeyboardInterrupt,
    PermissionError,
    TimeoutError,
    ConnectionError,
    ValueError,
    SystemError,
    SystemExit,
)



"""Setup the python module."""
from setuptools import setup, find_packages  # type: ignore

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = [
    "aiohttp<=3.8.4,>=3.7.4",
    "dicom2nifti<=2.4.7",
    "inquirerpy<=0.3.4,>=0.3.3",
    "natsort<=8.2.0,>=8.0.2",
    "nest-asyncio<=1.5.6,>=1.5.4",
    "nibabel<=5.0.1,>=3.2.2",
    "numpy<=1.24.2,>=1.15",
    "packaging<=23.0",
    "Pillow<=9.4.0,>=9.0.1",
    "python-dateutil<=2.8.2",
    "requests<=2.28.2,>=2.23.0",
    "Rich<=13.3.1",
    "shtab<=1.5.8",
    "tenacity<=8.2.1",
    "tqdm<=4.64.1,>=4.50.0",
]

setup(
    name="redbrick-sdk",
    url="https://github.com/redbrick-ai/redbrick-sdk",
    description="RedBrick platform Python SDK!",
    py_modules=["redbrick"],
    python_requires=">=3.8",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        "console_scripts": ["redbrick=redbrick.cli:cli_main"],
    },
    extras_require={
        "dev": [
            "twine==4.0.2",
            "wheel==0.38.4",
            "pytest==7.2.0",
            "black==22.12.0",
            "mypy==0.991",
            "mypy-extensions==0.4.3",
            "pylint==2.15.9",
            "pycodestyle==2.8.0",
            "pydocstyle==6.2.2",
            "flake8==4.0.1",
            "flake8-print==5.0.0",
            "eradicate==1.0",
            "pytest-cov>=4.0.0",
            "pytest-asyncio==0.20.3",
            "pytest-eradicate==0.0.5",
            "pytest-pycodestyle==2.3.1",
            "pytest-pydocstyle==2.3.2",
            "pytest-black==0.3.12",
            "pytest-pylint==0.19.0",
            "pytest-mypy==0.10.3",
            "pytest-flake8==1.1.1",
            "pytest-randomly==3.12.0",
            "pytest-xdist==3.1.0",
            "pytest-xdist[psutil]==3.1.0",
        ]
    },
)



"""Abstract interface to upload."""

from typing import List, Dict, Optional, Any
from abc import ABC, abstractmethod

import aiohttp


class UploadControllerInterface(ABC):
    """Abstract interface to define methods for Upload."""

    @abstractmethod
    async def create_datapoint_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        name: str,
        items: List[str],
        labels_data: Optional[str],
        labels_map: Optional[List[Dict]] = None,
        series_info: Optional[List[Dict]] = None,
        meta_data: Optional[str] = None,
        is_ground_truth: bool = False,
        pre_assign: Optional[Dict] = None,
    ) -> Dict:
        """
        Create a datapoint and returns its taskId.

        Name must be unique in the project.
        """

    @abstractmethod
    async def update_items_async(
        self,
        aio_client: aiohttp.ClientSession,
        org_id: str,
        project_id: str,
        storage_id: str,
        task_id: str,
        items: List[str],
        series_info: Optional[List[Dict]] = None,
    ) -> Dict:
        """Update items in a datapoint."""

    @abstractmethod
    def items_upload_presign(
        self, org_id: str, project_id: str, files: List[str], file_type: List[str]
    ) -> List[Dict[Any, Any]]:
        """Get a presigned url for uploading items."""

    @abstractmethod
    def delete_tasks(self, org_id: str, project_id: str, task_ids: List[str]) -> bool:
        """Delete tasks in a project."""

    @abstractmethod
    def delete_tasks_by_name(
        self, org_id: str, project_id: str, task_names: List[str]
    ) -> bool:
        """Delete tasks in a project by task names."""

    @abstractmethod
    async def generate_items_list(
        self,
        aio_client: aiohttp.ClientSession,
        files: List[str],
        import_type: str,
        as_study: bool = False,
        windows: bool = False,
    ) -> str:
        """Generate direct upload items list."""

    @abstractmethod
    async def validate_and_convert_to_import_format(
        self,
        aio_client: aiohttp.ClientSession,
        original: str,
        convert: Optional[bool] = None,
        storage_id: Optional[str] = None,
    ) -> Dict:
        """Validate and convert tasks format."""

    @abstractmethod
    def import_tasks_from_workspace(
        self,
        org_id: str,
        project_id: str,
        source_project_id: str,
        task_search: List[Dict],
        with_labels: bool = False,
    ) -> Dict:
        """Import tasks from another project in the same workspace."""
